import 'dart:convert';
import 'dart:ffi';

import 'package:flutter/services.dart';

import 'kameleoon_data.dart';
import 'kameleoon_exceptions.dart';

class KameleoonClientCallback {
  Function readyCallback;
  Function failCallback;
  int expires;

  KameleoonClientCallback(Function readyCallback, Function failCallback, int expires) {
    this.readyCallback = readyCallback;
    this.failCallback = failCallback;
    this.expires = expires;
  }
}

class KameleoonClient {
  static const MethodChannel _channel = const MethodChannel("kameleoon_client_flutter");
  bool ready = false;
  List<KameleoonClientCallback> callbacks = [];
  String siteCode;

  static final Map<String, KameleoonClient > _clients = {};

  KameleoonClient._createInstance(this.siteCode) {
    _channel.setMethodCallHandler(nativeMethodCallHandler);
    _channel.invokeMethod("initializeKameleoonClient", <String, dynamic>{
      "siteCode": this.siteCode
    });
  }

  factory KameleoonClient(String siteCode) {
    if (!_clients.containsKey(siteCode)) {
      var instance = KameleoonClient._createInstance(siteCode);
      _clients[siteCode] = instance;
    }
    return _clients[siteCode];
  }

  static Future<KameleoonClient> create(String siteCode) async {
    if (!_clients.containsKey(siteCode)) {
      var instance = KameleoonClient._createInstance(siteCode);
      _clients[siteCode] = instance;
    }
    return _clients[siteCode];
  }

  Future<void> trackConversion(String visitorCode, int goalID, [Float revenue]) async {
    await _channel.invokeMethod("trackConversion", <String, dynamic>{
      "visitorCode": visitorCode,
      "goalID": goalID,
      "revenue": revenue,
      "siteCode": this.siteCode
    });
  }

  Future<int> triggerExperiment(String visitorCode, int experimentID) async {
    try {
      return await _channel.invokeMethod("triggerExperiment", <String, dynamic>{
        "visitorCode": visitorCode,
        "experimentID": experimentID,
        "siteCode": this.siteCode
      });
    } on PlatformException catch(e) {
      switch (e.code) {
        case "SDKNotReady":
          throw SDKNotReady();
        case "ExperimentConfigurationNotFound":
          throw ExperimentConfigurationNotFound();
        case "NotTargeted":
          throw NotTargeted();
        case "NotActivated":
          throw NotActivated();
        default:
          throw Exception(e);
      }
    }
  }

  Future<bool> activateFeature(String visitorCode, var featureKeyOrID) async {
    try {
      if (featureKeyOrID is int) {
        return await _channel.invokeMethod("activateFeature", <String, dynamic>{
          "visitorCode": visitorCode,
          "featureID": featureKeyOrID,
          "siteCode": this.siteCode
        });
      }
      else {
        return await _channel.invokeMethod("activateFeature", <String, dynamic>{
          "visitorCode": visitorCode,
          "featureKey": featureKeyOrID,
          "siteCode": this.siteCode
        });
      }
    } on PlatformException catch(e) {
      switch (e.code) {
        case "SDKNotReady":
          throw SDKNotReady();
        case "FeatureConfigurationNotFound":
          throw FeatureConfigurationNotFound();
        case "NotTargeted":
          throw NotTargeted();
        case "NotActivated":
          throw NotActivated();
        default:
          throw Exception(e);
      }
    }
  }

  Future<String> obtainVisitorCode([String defaultVisitorCode]) async {
    return await _channel.invokeMethod("obtainVisitorCode", <String, dynamic>{
      "defaultVisitorCode": defaultVisitorCode,
      "siteCode": this.siteCode
    });
  }

  Future<dynamic> obtainVariationAssociatedData(int variationID) async {
    try {
      String objAsString = await _channel.invokeMethod("obtainVariationAssociatedData", <String, dynamic>{
        "variationID": variationID,
        "siteCode": this.siteCode
      });
      return jsonDecode(objAsString);
    } on PlatformException catch(e) {
      switch (e.code) {
        case "VariationConfigurationNotFound":
          throw VariationConfigurationNotFound();
        default:
          throw Exception(e);
      }
    }
  }

  Future<Object> obtainFeatureVariable(var featureKeyOrID, String variableKey) async {
    if (featureKeyOrID is int) {
      return await _channel.invokeMethod("obtainFeatureVariable", <String, dynamic>{
        "variableKey": variableKey,
        "featureID": featureKeyOrID,
        "siteCode": this.siteCode
      });
    }
    else {
      return await _channel.invokeMethod("obtainFeatureVariable", <String, dynamic>{
        "variableKey": variableKey,
        "featureKey": featureKeyOrID,
        "siteCode": this.siteCode
      });
    }
  }

  Future<void> addData(String visitorCode, Data data) async {
    var parameters = <String, dynamic>{};
    parameters.putIfAbsent("visitorCode", () => visitorCode);
    parameters.putIfAbsent("dataType", () => data.dataType);
    parameters.putIfAbsent("siteCode", () => this.siteCode);
    switch (data.dataType) {
      case "CustomData":
        CustomData d = data;
        parameters.putIfAbsent("id", () => d.id);
        parameters.putIfAbsent("value", () => d.value);
        break;
      case "PageView":
        PageView d = data;
        parameters.putIfAbsent("url", () => d.url);
        parameters.putIfAbsent("title", () => d.title);
        parameters.putIfAbsent("referrer", () => d.referrer);
        break;
      case "Interest":
        Interest d = data;
        parameters.putIfAbsent("index", () => d.index);
        break;
      case "Conversion":
        Conversion d = data;
        parameters.putIfAbsent("goalId", () => d.goalId);
        parameters.putIfAbsent("revenue", () => d.revenue);
        parameters.putIfAbsent("negative", () => d.negative);
        break;
      case "Browser":
        Browser d = data;
        parameters.putIfAbsent("browser", () => d.browser);
        break;
    }
    await _channel.invokeMethod("addData", parameters);
  }

  void runWhenReady(Function readyCallback, Function failCallback, [int timeout = 2000]) async {
    if (this.ready) {
      await readyCallback.call();
    } else {
      this.callbacks.add(new KameleoonClientCallback(readyCallback, failCallback, 2000));
      await _channel.invokeMethod("runWhenReady", <String, dynamic>{
        "timeout": timeout,
        "siteCode": this.siteCode
      });
    }
  }

  void callbackAfterStatusChange(bool isReady, [KameleoonClientCallback callback]) {
    List<int> idsToDelete = [];
    for (int i = 0; i < this.callbacks.length; i++) {
      if (isReady) {
        this.callbacks[i].readyCallback();
        idsToDelete.add(i);
      }
    }

    for (int i = idsToDelete.length - 1; i >= 0; i--) {
      this.callbacks.removeAt(i);
    }
  }

  Future<void> nativeMethodCallHandler(MethodCall call) async {
    switch (call.method) {
      case "setReady" :
        this.ready = call.arguments;
        callbackAfterStatusChange(this.ready);
        break;
      default:
        break;
    }
  }
}